//serial helper functions for the Poisson equation
//none of these are related to CnC
#include "p2.h"

double serial_dot(double *x, double *y, int n){
  double d = 0.0;
  int i;
  for(i=0; i < n; i++){
    d += x[i] * y[i];
  }
  return d;
}

void Ax(double *v, double *x, int N, double h, double p)
{
  double temp;
  int i, j;

  for (j = 0; j < N; j ++){
    for (i = 0; i < N; i ++){
      temp = (4.0 + (h*h)*p) * x[i + N*j];
      if (j > 0){
	temp -= x[i + N*(j-1)];
	  }
      if (i > 0){
	temp -= x[(i-1) + N*j];
	  }
      if (i < (N-1)){
	temp -= x[(i+1) + N*j];
      }
      if (j < (N-1)){
	temp -= x[i + N*(j+1)];
      }
      v[i + N*j] = temp;
    }
  }

}

void cg(double *u, int *flag, double *relres, int *iter, /* output */
        double *l_r, double tol, int maxit, /* input */
        double *l_p, double *l_q,
        int l_n, int l_N, int N, double h, double p) {
  
  int it, l_i;
  double n2b, tolb, normr, alpha, pq, beta;
  double rho = 0.0; 
  double rho1 = 0.0;

  n2b = sqrt(serial_dot(l_r,l_r,l_n)); /* Norm of the rhs vector b */
  if(n2b <= 1.0e-14) {           /* if the rhs vector is all zeros...     */
    for (l_i=0; l_i<l_n; l_i++)  /* then the solution is all zeros        */
      u[l_i] = 0.0;
    *flag = 0;                   /* a valid solution has been obtained    */
    *relres = 0;                 /* the relative residual is actually 0/0 */
    *iter = 0;                   /* no iterations need to be performed    */
    return;
  }

  tolb = tol * n2b; /* relative tolerance */
  Ax(l_q, u, N, h, p); /* q = A * x   */
  for (l_i=0; l_i<l_n; l_i++)                        /* r = r - A*x       */
    l_r[l_i] = l_r[l_i] - l_q[l_i];
  rho = serial_dot(l_r, l_r, l_n);           /* rho = r' * r      */
  normr = sqrt(rho);                                 /* normr = norm(r)   */
  
  /* initial guess is a good enough: */
  if(normr <= tolb) {
    *flag = 0;
    *relres = normr / n2b;
    *iter = 0;
    return;
  }

  it = 0;
  while ((normr > tolb) && (it < maxit)) {
    it++;

    if(it == 1) {
      for (l_i=0; l_i<l_n; l_i++)                    /* p = r             */
        l_p[l_i] = l_r[l_i];
    } else {
      beta = rho / rho1;
      for (l_i=0; l_i<l_n; l_i++)                    /* p = r + beta * p  */
        l_p[l_i] = l_r[l_i] + beta * l_p[l_i];
    }

    Ax(l_q, l_p, N, h, p); /* q = A * p */

    pq = serial_dot(l_p, l_q, l_n);          /* pq = p' * q;      */
    alpha = rho / pq;                                /* alpha = rho/p'*q  */

    for (l_i=0; l_i<l_n; l_i++)                      /* r = r - alpha * q */
      l_r[l_i] = l_r[l_i] - alpha * l_q[l_i];
    rho1 = rho;                                      /* rho1 = rho;       */
    rho = serial_dot(l_r, l_r, l_n);         /* rho = r' * r */
    normr = sqrt(rho);                               /* normr = norm(r)   */

    for (l_i=0; l_i<l_n; l_i++)                      /* x = x + alpha * p */
      u[l_i] = u[l_i] + alpha * l_p[l_i];
  }

  /* when the loop completes it will break to here - set output variables */  
  if (it < maxit)
    *flag = 0;
  else
    *flag = 1; /* maxit iterations reached */
  *relres = normr / n2b;
  *iter = it;

}

void poisson(int N, double tol, int maxit, double *maxError, int *iter, 
	     double a)
{


  int n, flag, i, j;
  double relres;

  n = N * N;
  relres = 0.0; 

  double h = 1.0 / ((double)(N + 1));
  double *u = (double *)calloc(n, sizeof(double));
  double *r = (double *)calloc(n, sizeof(double));
  double *p = (double *)calloc(n, sizeof(double));
  double *q = (double *)calloc(n, sizeof(double));
  //x & y are the mesh points of the arrays
  double *x = (double *)calloc(N, sizeof(double));
  double *y = (double *)calloc(N, sizeof(double));

  //for loops that fill the x and y arrays
  for (i = 0; i < N; i++) 
    {
      x[i] = h * ((double)(i + 1));
    }

  for (j = 0; j < N; j ++)
    {
      y[j] = h * ((double)(j + 1));
    }

  //actual poisson function's left side, run with local N by id
  double PI = 4.0*atan(1.0);

  for (j = 0; j < N; j ++)
    {
      for (i = 0; i < N; i ++)
	{
	  r[i + N*j] = (h*h) * (
              (-2.0 * PI * PI) * 
	      (cos(2.0 * PI * x[i]) * pow(sin(PI * y[j]), 2) + 
	       pow(sin(PI * x[i]), 2) * cos(2.0 * PI * y[j]))
            + a * pow(sin(PI*x[i]),2) * pow(sin(PI*y[j]),2)
                               );
	}
    }

  /*wait*/
  
  int iter1 = 0;
  //call to the cg function that finishes the poisson function
  cg(u, &flag, &relres, &iter1, r, tol, maxit, p, q, n, N, N, h, a);
 
  //computes the difference (error) between U(u) and UTrue(r2)
  double temp = 0.0;
  double enorminf = 0.0;
  for (j = 0; j < N; j ++)
    {
      for (i = 0; i < N; i ++)
	{
          temp = fabs(u[i+N*j]-pow(sin(PI*x[i]),2) * pow(sin(PI*y[j]),2));
          if (temp > enorminf)
	    {
	      enorminf = temp;
	    }
        }
    }
  *iter = iter1;
  *maxError = enorminf;

  free(u);
  free(r);
  free(p);
  free(q);
  free(x);
  free(y);

}
